/***********************************************************************
Filename: Event.c
***********************************************************************/
#define _EVENT_C_

#include <mrt.h>
#include <config_file.h>
#include <stdlib.h>
#include <string.h>
#include <values.h>
#include "Event.h"
#include "Globals.h"

#define MIN_INTERVAL	30

#define MIN_PACKET_SIZE		1
#define MAX_PACKET_SIZE		1048576
#define DEFAULT_PACKET_SIZE	1024

#define MIN_TTL			1
#define DEFAULT_TTL		32
#define MAX_TTL			255

#define MAX_ID		999999

char	COMMAND_CONFIG_ADD[]	= {"add"};
char	COMMAND_CONFIG_REMOVE[]	= {"remove"};

/***********************************************************************
Create a new event and allocate all memory associated with it.
***********************************************************************/
void
CreateNewEvent(char *host, int interval, int PacketSize, int TTL) {

	TPDEvent	*NewEventPtr,
			*EventPtr;
	prefix_t	*NewPrefix;
	struct timeval	CurrentTime;
	int		ID;

	NewPrefix = string_toprefix(host, TPDGlobals.TracePtr);
	if (!NewPrefix)
		return;

	gettimeofday(&CurrentTime, NULL);

        NewEventPtr = (TPDEvent *)calloc(1, sizeof(TPDEvent));
        NewEventPtr->Next = NULL;
        NewEventPtr->NextStartTime.tv_sec     = CurrentTime.tv_sec;
        NewEventPtr->NextStartTime.tv_usec    = CurrentTime.tv_usec;
        NewEventPtr->Interval.tv_sec  = interval;
        NewEventPtr->Interval.tv_usec = 0;
	NewEventPtr->PacketSize = PacketSize;
	NewEventPtr->TTL  = TTL;
	NewEventPtr->Prefix = NewPrefix;
	NewEventPtr->DstAddr = host;

	GetEventListLock();

	do {
		ID = rand() % MAX_ID;
		EventPtr = TPDGlobals.EventListHead.Next;
		while (EventPtr) {
			if (EventPtr->ID == ID)
				break;
			EventPtr = EventPtr->Next;
		}
	} while (EventPtr);
	NewEventPtr->ID = ID;
	NewEventPtr->Running = 0;

        AddEvent(NewEventPtr);
	ReleaseEventListLock();
}


/***********************************************************************
Destroy an event and free all memory associated with it.
***********************************************************************/
void
DestroyEvent(TPDEvent *EventPtr) {
	Delete_Prefix(EventPtr->Prefix);
	free(EventPtr->DstAddr);
	free(EventPtr);
}

/***********************************************************************
Add the given event to the global set of pending events.
***********************************************************************/
void
AddEvent(TPDEvent *EventPtr) {
	EventPtr->Next			= TPDGlobals.EventListHead.Next;
	TPDGlobals.EventListHead.Next	= EventPtr;
}

/***********************************************************************
Look for the specified event and remove it if it exists.
***********************************************************************/
void
RemoveEvent(TPDEvent * EventPtr) {
	TPDEvent	*PrevPtr;

	// Start at Head node
	PrevPtr = &(TPDGlobals.EventListHead);

	//Traverse linked list until we find it
	while (PrevPtr->Next) {
		if (PrevPtr->Next == EventPtr) {
			PrevPtr->Next = EventPtr->Next;
			break;
		}
		else {
			PrevPtr = PrevPtr->Next;
		}
	}
}

/***********************************************************************
Obtain the next waiting event, if there is one.
***********************************************************************/
TPDEvent *
GetNextEvent(void) {
	TPDEvent		*EventPtr,
				*NextPtr;
	struct timeval		NextTime;

	NextPtr			= NULL;
	NextTime.tv_sec		= MAXLONG;
	NextTime.tv_usec	= MAXLONG;
	EventPtr		= TPDGlobals.EventListHead.Next;

	while (EventPtr) {
		if (EventPtr->NextStartTime.tv_sec < NextTime.tv_sec) {
			NextPtr			= EventPtr;
			NextTime.tv_sec		= EventPtr->NextStartTime.tv_sec;
			NextTime.tv_usec	= EventPtr->NextStartTime.tv_usec;
		}
		EventPtr = EventPtr->Next;
	}

	return NextPtr;
}


/***********************************************************************
Add a traceroute to the set of pending events.
***********************************************************************/
int
AddTraceroute(uii_connection_t *uii, char *DstAddr, int Interval) {

	prefix_t	*NewPrefix;

	NewPrefix = string_toprefix(DstAddr, TPDGlobals.TracePtr);

	if (!NewPrefix) {
		config_notice(NORM, uii,
			"Couldn't find a proper prefix\r\n");
		free(DstAddr);
		return 0xbad;
	}
	Delete_Prefix(NewPrefix);

	if (Interval < MIN_INTERVAL) {
		config_notice(NORM, uii,
			"Please choose a bigger time interval\r\n");
		Delete_Prefix(NewPrefix);
		free(DstAddr);
		return 0xbadace;
	}

	// If we get here, that means we have a valid entry
	CreateNewEvent(DstAddr, Interval, DEFAULT_PACKET_SIZE,
		DEFAULT_TTL);
	trace (NORM, TPDGlobals.TracePtr,
		"Adding tpd host %s\n", DstAddr);
	config_notice(NORM, uii,
		"Created new entry for %s %d\r\n",
		DstAddr, Interval);

	return 0xbad;
}


/***********************************************************************
Add a traceroute to the set of pending events.
***********************************************************************/
int
AddTracerouteFull(uii_connection_t *uii, char *DstAddr, int Interval,
	int PacketSize, int TTL) {

	prefix_t	*NewPrefix;

	NewPrefix = string_toprefix(DstAddr, TPDGlobals.TracePtr);

	if (!NewPrefix) {
		config_notice(NORM, uii,
			"Couldn't find a proper prefix\r\n");
		free(DstAddr);
		return 0xbad;
	}
	Delete_Prefix(NewPrefix);

	if (Interval < MIN_INTERVAL) {
		config_notice(NORM, uii,
			"Please choose a bigger time interval\r\n");
		Delete_Prefix(NewPrefix);
		free(DstAddr);
		return 0xbadace;
	}

	if ((PacketSize < MIN_PACKET_SIZE) ||
	    (PacketSize > MAX_PACKET_SIZE)) {
		config_notice(NORM, uii,
			"Please choose a size from %d to %d\r\n",
			MIN_PACKET_SIZE, MAX_PACKET_SIZE);
		Delete_Prefix(NewPrefix);
		free(DstAddr);
		return 0xbadacc;
	}

	if ((TTL < MIN_TTL) || (TTL > MAX_TTL)) {
		config_notice(NORM, uii,
			"Please choose a TTL from %d to %d\r\n",
			MIN_TTL, MAX_TTL);
		Delete_Prefix(NewPrefix);
		free(DstAddr);
		return 0xbadacc;
	}

	// If we get here, that means we have a valid entry
	CreateNewEvent(DstAddr, Interval, PacketSize, TTL);
	trace (NORM, TPDGlobals.TracePtr,
		"Adding tpd host %s\n", DstAddr);
	config_notice(NORM, uii,
		"Created new entry for %s %d\r\n",
		DstAddr, Interval);

	return 0xbad;
}


/***********************************************************************
***********************************************************************/
int
ModifyTracerouteInterval(uii_connection_t *uii, int ID, int interval) {

	TPDEvent	*EventPtr;

	if (interval < MIN_INTERVAL) {
		config_notice(ERROR, uii,
			"Please choose an interval greater than %d\r\n",
			MIN_INTERVAL);
		return 0xbad;
	}

	if ((ID < 0) || (ID >= MAX_ID)) {
		config_notice(ERROR, uii,
			"ID %06d is invalid\r\n",
			ID);
		return 0xbad;
	}
	else {
		config_notice(NORM, uii,
			"Modifying ID %06d\r\n",
			ID);
	}

	GetEventListLock();

	EventPtr = TPDGlobals.EventListHead.Next;
	while (EventPtr) {
		if (EventPtr->ID == ID)
			break;
		else
			EventPtr = EventPtr->Next;
	}

	if (!EventPtr) {
		//Event doesn't exist
		config_notice(NORM, uii,
			"ID %06d does not exist\r\n",
			ID);
	}
	else {
		EventPtr->Interval.tv_sec = interval;
		config_notice(NORM, uii,
			"ID %06d was modified\r\n",
			ID);
	}

	ReleaseEventListLock();

	return 0;
}




/***********************************************************************
***********************************************************************/
int
ModifyTracerouteSize(uii_connection_t *uii, int ID, int PacketSize) {

	TPDEvent	*EventPtr;

	if ((PacketSize < MIN_PACKET_SIZE) ||
	    (PacketSize > MAX_PACKET_SIZE)) {
		config_notice(ERROR, uii,
			"Please choose a packet size from %d to %d\r\n",
			MIN_PACKET_SIZE, MAX_PACKET_SIZE);
		return 0xbad;
	}

	if ((ID < 0) || (ID >= MAX_ID)) {
		config_notice(ERROR, uii,
			"ID %06d is invalid\r\n",
			ID);
		return 0xbad;
	}
	else {
		config_notice(NORM, uii,
			"Modifying ID %06d\r\n",
			ID);
	}

	GetEventListLock();

	EventPtr = TPDGlobals.EventListHead.Next;
	while (EventPtr) {
		if (EventPtr->ID == ID)
			break;
		else
			EventPtr = EventPtr->Next;
	}

	if (!EventPtr) {
		//Event doesn't exist
		config_notice(NORM, uii,
			"ID %06d does not exist\r\n",
			ID);
	}
	else {
		EventPtr->PacketSize = PacketSize;
		config_notice(NORM, uii,
			"ID %06d was modified\r\n",
			ID);
	}

	ReleaseEventListLock();

	return 0;
}


/***********************************************************************
***********************************************************************/
int
ModifyTracerouteTTL(uii_connection_t *uii, int ID, int TTL) {

	TPDEvent	*EventPtr;

	if ((TTL < MIN_TTL) || (TTL > MAX_TTL)) {
		config_notice(ERROR, uii,
			"Please choose a TTL from %d to %d\r\n",
			MIN_TTL, MAX_TTL);
		return 0xbad;
	}

	if ((ID < 0) || (ID >= MAX_ID)) {
		config_notice(ERROR, uii,
			"ID %06d is invalid\r\n",
			ID);
		return 0xbad;
	}
	else {
		config_notice(NORM, uii,
			"Removing ID %06d\r\n",
			ID);
	}

	GetEventListLock();

	EventPtr = TPDGlobals.EventListHead.Next;
	while (EventPtr) {
		if (EventPtr->ID == ID)
			break;
		else
			EventPtr = EventPtr->Next;
	}

	if (!EventPtr) {
		//Event doesn't exist
		config_notice(NORM, uii,
			"ID %06d does not exist\r\n",
			ID);
	}
	else {
		EventPtr->TTL = TTL;
		config_notice(NORM, uii,
			"ID %06d was found\r\n",
			ID);
	}

	ReleaseEventListLock();

	return 0;
}


/***********************************************************************
Remove a traceroute event with the specified ID from the set of events.
***********************************************************************/
int
RemoveTraceroute(uii_connection_t *uii, int ID) {

	TPDEvent	*EventPtr;

	if ((ID < 0) || (ID >= MAX_ID)) {
		config_notice(ERROR, uii,
			"ID %06d is invalid\r\n",
			ID);
		return 0xbad;
	}
	else {
		config_notice(NORM, uii,
			"Removing ID %06d\r\n",
			ID);
	}

	GetEventListLock();

	EventPtr = TPDGlobals.EventListHead.Next;
	while (EventPtr) {
		if (EventPtr->ID == ID) {
			RemoveEvent(EventPtr);
			break;
		}
		EventPtr = EventPtr->Next;
	}

	if (!EventPtr) {
		//Event doesn't exist
		config_notice(NORM, uii,
			"ID %06d does not exist\r\n",
			ID);
	}
	else {
		config_notice(NORM, uii,
			"ID %06d was removed\r\n",
			ID);
		if (EventPtr->Running > 0) {
			// We found it, but it is currently running
			EventPtr->Interval.tv_sec  = 0xffffff;
			EventPtr->Interval.tv_usec = 0xffffff;
			EventPtr->Running = -EventPtr->Running;
		}
		else if (EventPtr->Running < 0) {
			// We found it, and it is being deleted later
			// so leave it alone
			;
		}
		else {
			// We found it, and can nuke it
			DestroyEvent(EventPtr);
		}
	}

	ReleaseEventListLock();

	return 0;
}


/***********************************************************************
Output the current set of traceroute events to the uii
***********************************************************************/
int
ShowTraceroutes(uii_connection_t *uii) {

        TPDEvent        *TPDEventPtr;

        GetEventListLock();
        TPDEventPtr = TPDGlobals.EventListHead.Next;
        if (TPDEventPtr) {
                do {
			config_notice(INFO, uii,
				"%06d: %s %d %d %d\r\r\n",
				TPDEventPtr->ID,
				TPDEventPtr->DstAddr,
				TPDEventPtr->Interval.tv_sec,
				TPDEventPtr->PacketSize,
				TPDEventPtr->TTL);
                        TPDEventPtr = TPDEventPtr->Next;
                } while (TPDEventPtr);
        }

        ReleaseEventListLock();
	uii_send_bulk_data (uii);
	return 0;
}

/***********************************************************************
Output the current set of traceroute events to the uii
***********************************************************************/
void
GetTracerouteConfig(void) {

        TPDEvent        *TPDEventPtr;

        GetEventListLock();
        TPDEventPtr = TPDGlobals.EventListHead.Next;
        if (TPDEventPtr) {
                do {
	        	config_add_output (
				"traceroutefull %s %d %d %d\r\n",
				TPDEventPtr->DstAddr,
				TPDEventPtr->Interval.tv_sec,
				TPDEventPtr->PacketSize,
				TPDEventPtr->TTL);
                        TPDEventPtr = TPDEventPtr->Next;
                } while (TPDEventPtr);
        }

        ReleaseEventListLock();
	return;
}

/***********************************************************************
***********************************************************************/
void
get_config_tpd_traceroute() {
  TPDEvent        *TPDEventPtr;

  GetEventListLock();
  TPDEventPtr = TPDGlobals.EventListHead.Next;
  if (TPDEventPtr) {
    do {
        config_add_output (
		"traceroutefull %s %d %d %d\r\n",
		TPDEventPtr->DstAddr,
		TPDEventPtr->Interval.tv_sec,
		TPDEventPtr->PacketSize,
		TPDEventPtr->TTL);
	TPDEventPtr = TPDEventPtr->Next;
    } while (TPDEventPtr);
  }

  ReleaseEventListLock();
}
